"""
MIT License

Copyright (c) 2017 Dominik Kopczynski   -   dominik.kopczynski {at} isas.de
                   Bing Peng   -   bing.peng {at} isas.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


####################################################################################
####### Please do NOT change anything, unless you know what you do!
####################################################################################

import time
import sys
import matplotlib.pyplot as plt
import matplotlib.colors as cls
from matplotlib import lines
import numpy as np
import random
from math import pi, sin, cos, sqrt, exp, log, floor
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.patches import Rectangle
from scipy.spatial.distance import pdist, squareform
from scipy.cluster import hierarchy
from scipy import linalg as LA
import time
import datetime
from lipid_parameters import *
import matplotlib.patches as mpatches

cdict_correlation = {'red':  ((0.0, 0.0, 0.0), (0.5, 1.0, 1.0), (1.0, 0.6, 0.6)),
          'green': ((0.0, 0.44, 0.44), (0.5, 1.0, 1.0), (1.0, 0.0, 0.0)),
          'blue':  ((0.0, 0.67, 0.67), (0.5, 1.0, 1.0), (1.0, 0.0, 0.0))}

colors_correlation = LinearSegmentedColormap('BlueRedCorrelation', cdict_correlation)

if len(sys.argv) < 3:
    print("usage:", sys.argv[0], "input-csv-file output-folder")
    exit()
    
    
input_csv = sys.argv[1]
output_folder = sys.argv[2]
intensities = []
lipids = []
classes = []
fatty_acids = []
fa_min, fa_max = 1000, -1000
column_names = ""
num_columns = -1
folder_sep = "\\" if sys.platform[:3] == "win" else "/"

time_stamp = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
print("%s \t\t reading '%s' as input file" % (time_stamp, input_csv))

with open(input_csv, mode = "rt") as input_csv_file:
    column_names = [name.strip(" ").replace(" ", "_") for name in input_csv_file.readline().strip().split(",")[2:]] # first line contains column names
    num_columns = len(column_names)
    for line_i, line in enumerate(input_csv_file):
        tokens = [name.strip(" ") for name in filter(bool, line.strip().split(csv_delimiter))]
        if (len(tokens) & 1) == 1: 
            print(tokens)
            continue
        try:
            if len(tokens) != 2 + num_columns:
                print("error in line %i, number of columns unequal to number of column names in first row" % (line_i + 1))
                exit()
            intensities.append([float(x) for x in tokens[2:]])
            
            lipids.append(tokens[1].replace(" ", "_"))
            if tokens[0] not in color_map:
                print("error in line %i, lipid category not registered in color map 'color_map'. Please consider the parameters section in the script" % (line_i + 1))
                exit()
            classes.append(tokens[0])
            fatty_acids.append([])
            for fa in range(2 + num_columns, len(tokens), 2):
                fatty_acids[-1].append([int(tokens[fa]), int(tokens[fa + 1])])
                fa_min = min(fa_min,fatty_acids[-1][-1][0])
                fa_max = max(fa_max,fatty_acids[-1][-1][0])
        except:
            print("error in line %i" % (line_i + 1))
            continue
        

time_stamp = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
print("%s \t\t input was successful" % time_stamp)
class_names = list(set(classes))
lipid_clusterings = [""] * len(lipids)
class_dict = {name: i for i, name in enumerate(class_names)}
n = len(intensities)

intensities_np = np.matrix(intensities, dtype = "float64")
clusters = []

time_stamp = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
print("%s \t\t computing pair wise Pearson correlations for each lipid profile pair" % time_stamp)
edge_rank = [[0, i, [], True] for i in range(n)]
correlations = pdist(intensities_np, 'correlation')
correlations = squareform(correlations)
correlations = 1 - correlations
pca_data = correlations.copy()

for i in range(n - 1):
    for j in range(i + 1, n):
        if correlations[i, j] <= -tolerance or tolerance <= correlations[i, j]:
            edge_rank[i][0] += 1
            edge_rank[i][2].append(j)
            edge_rank[j][0] += 1
            edge_rank[j][2].append(i)


edge_rank.sort(reverse = True)
max_cluster = edge_rank[0][0]



def principal_component_analysis(data, dims_rescaled_data=2):
    """
    returns: data transformed in 2 dims/columns + regenerated original data
    pass in: data as 2D NumPy array
    """
    m, n = data.shape
    # mean center the data
    data -= data.mean(axis = 0)
    # calculate the covariance matrix
    R = np.cov(data, rowvar = False)
    # calculate eigenvectors & eigenvalues of the covariance matrix
    # use 'eigh' rather than 'eig' since R is symmetric, 
    # the performance gain is substantial
    evals, evecs = LA.eigh(R)
    # sort eigenvalue in decreasing order
    idx = np.argsort(evals)[::-1]
    evecs = evecs[:,idx]
    # sort eigenvectors according to same index
    evals = evals[idx]
    # select the first n eigenvectors (n is desired dimension
    # of rescaled data array, or dims_rescaled_data)
    evecs = evecs[:, :dims_rescaled_data]
    # carry out the transformation on the data using eigenvectors
    # and return the re-scaled data, eigenvalues, and eigenvectors
    return np.dot(evecs.T, data.T).T, evals, evecs


time_stamp = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
print("%s \t\t computing principal component analysis" % time_stamp)
pca_data = principal_component_analysis(np.abs(pca_data))[0] if with_anticorrelation else principal_component_analysis(pca_data)[0]
pca_data *= 3


min_x, max_x, min_y, max_y = 1e5, -1e5, 1e5, -1e5
for i in range(n):
    x, y = pca_data[i,0], pca_data[i,1]
    min_x = min(min_x, x)
    max_x = max(max_x, x)
    min_y = min(min_y, y)
    max_y = max(max_y, y)
min_x -= 0.5
max_x += 0.5
min_y -= 0.5
max_y += 0.5

non_singleton = [False] * n
for i in range(n - 1):
    for j in range(i + 1, n):
        if correlations[i, j] < -tolerance or tolerance < correlations[i, j]: non_singleton[i], non_singleton[j] = True, True
   
##################################################################################
# create gml
##################################################################################


if create_gml:
    time_stamp = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
    print("%s \t\t writing gml overview" % time_stamp)
    with open("%s%snetwork-graph.gml" % (output_folder, folder_sep), mode = "wt") as gml:
        gml.write("graph [\n")
        gml.write("\tdirected 0\n")
        gml.write("\tid 0\n")
        gml.write("\tlabel \"Lipid network analysis\"\n")
        for i in range(n):
            gml.write("\tnode [\n")
            gml.write("\t\tid %i\n" % i)
            gml.write("\t\tlabel \"%s\"\n" % lipids[i])
            gml.write("\t\tcategory \"%s\"\n" % classes[i])
            gml.write("\t\tx %0.3f\n" % pca_data[i,0])
            gml.write("\t\ty %0.3f\n" % pca_data[i,1])
            mode = 0
            for mode2 in range(1, len(column_names) // 2):                
                col_name = column_names[mode2] + "_VS_" + column_names[mode]
                log_fold = 0
                try:
                    log_fold = intensities_np[i, mode2] / intensities_np[i, mode]
                except:
                    pass
                gml.write("\t\t%s %0.3f\n" % (col_name, log_fold))
                
            
            for mode in range(0, len(column_names) // 2):
                mode2 = mode + len(column_names) // 2
                col_name = column_names[mode2] + "_VS_" + column_names[mode]
                log_fold = 0
                try:
                    log_fold = intensities_np[i, mode2] / intensities_np[i, mode]
                except:
                    pass
                gml.write("\t\t%s %0.3f\n" % (col_name, log_fold))
                
                
            
            node_degree = 0
            for j in range(n):
                if correlations[i, j] <= -tolerance or tolerance <= correlations[i, j] and i != j: node_degree += 1
            node_degree = max(10, node_degree)
            node_degree = min(200, node_degree)
            node_degree *= 2
            
            size_abundance = intensities_np[i, 0]
            
            gml.write("\t\tsize_degree %0.3f\n" % node_degree)
            gml.write("\t\tsize_abundance %0.3f\n" % size_abundance)
            gml.write("\t\tcolor \"%s\"\n" % color_map[classes[i]])
            gml.write("\t]\n")
            
        for i in range(n - 1):
            for j in range(i + 1, n):
                if correlations[i, j] <= -tolerance or tolerance <= correlations[i, j]:
                    gml.write("\tedge [\n")
                    gml.write("\t\tsource %i\n" % i)
                    gml.write("\t\ttarget %i\n" % j)
                    gml.write("\t\tweight %0.4f\n" % correlations[i, j])
                    gml.write("\t]\n")
        gml.write("]\n")
    

##################################################################################
# correlation matrix and group figure
##################################################################################
    
    
    
if create_correlation_matrix:
    time_stamp = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
    print("%s \t\t correlation matrix: computing hierarchical clustering with average linkage and cosine similarity" % time_stamp)
    indexes = [[i] for i in range(n)]
    clustering = np.abs(correlations.copy()) if with_anticorrelation else correlations.copy()
    
    hierarchical_clustering = [[i] for i in range(len(clustering))]
    linkage = hierarchy.linkage(clustering, method="average", metric="cosine")
    for l in linkage:
        hierarchical_clustering.append([hierarchical_clustering[int(l[0])], hierarchical_clustering[int(l[1])]])
    
    order = []
    def rec(lst):
        if len(lst) == 1: order.append(lst)
        else:
            rec(lst[0])
            rec(lst[1])
    rec(hierarchical_clustering[-1])
    order = [x[0] for x in order]
    
    
    anti_order = [0] * n
    for i, j in enumerate(order): anti_order[j] = i
    
    sorted_correlations_abs = (np.abs(correlations)).tolist() if with_anticorrelation else (correlations).tolist()
    sorted_correlations = correlations.tolist()
    sorted_correlations = list(x for x, y in sorted(zip(sorted_correlations, anti_order), key = lambda x: x[1]))
    sorted_correlations_abs = list(x for x, y in sorted(zip(sorted_correlations_abs, anti_order), key = lambda x: x[1]))
    for i in range(n):
        sorted_correlations[i] = [x for x, y in sorted(zip(sorted_correlations[i], anti_order), key = lambda x: x[1])]
        sorted_correlations_abs[i] = [x for x, y in sorted(zip(sorted_correlations_abs[i], anti_order), key = lambda x: x[1])]
    sorted_correlations_np = np.matrix(sorted_correlations)
    sorted_correlations_np_abs = np.matrix(sorted_correlations_abs)
        
    # find cluster
    time_stamp = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
    print("%s \t\t searching for clusters, min cluster number: %i, max cluster number %i" % (time_stamp, min_cluster, max_cluster))
    best_i, best_c, hit = 0, 0, True
    used_i = set()
    cnt = 0
    while hit:
        hit = False
        for cluster_size in range(min_cluster, max_cluster + 1):
            for i in range(n - cluster_size + 1):
                sum_corr = 0
                go_on = True
                for j in range(i, i + cluster_size):
                    if j in used_i: 
                        go_on = False
                        break
                if go_on:
                    sum_corr = np.sum(sorted_correlations_np_abs[i:i + cluster_size, i: i + cluster_size]) / cluster_size**2
                    #sum_corr = np.min(sorted_correlations_np_abs[i:i + cluster_size, i: i + cluster_size])
                                    
                    if tolerance < sum_corr:
                        best_i, best_c = i, cluster_size
                        hit = True
        if hit:
            for i in range(best_i, best_i + best_c):
                used_i.add(i)
            clusters.append([best_i, best_c, [order[index] for index in range(best_i, best_i + best_c)]])
        cnt += 1
        
    for i in range(len(clusters)):
        clusters[i].append([0] * len(class_names))
        for lipid in clusters[i][2]:
            clusters[i][3][class_dict[classes[lipid]]] += 1
    
    max_bar_width = 0
    for i in range(len(clusters)):
        for j in range(1, len(class_names)):
            clusters[i][3][j] += clusters[i][3][j - 1]
            max_bar_width = max(max_bar_width, clusters[i][3][j])
            
    clusters.sort()
    time_stamp = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
    print("%s \t\t found %i clusters" % (time_stamp, len(clusters)))
    
    
    for i, cluster in enumerate(clusters):
        for index in cluster[2]:
            lipid_clusterings[index] = "C%i" % (i + 1)
            
    with open(output_folder + folder_sep + input_csv.split(folder_sep)[-1].replace(".csv", "_clustered.csv"), mode = "wt") as cluster_output:
        cluster_output.write("lipid;cluster\n")
        for lipid, lipid_clustering in zip(lipids, lipid_clusterings):
            cluster_output.write("%s;%s\n" % (lipid, lipid_clustering))
    
        
    
    time_stamp = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
    print("%s \t\t plotting matrix and groups" % time_stamp)
    fig, ax = plt.subplots(1, 2, sharey = True)
    fig.set_size_inches(8.7, 4)
    plt.subplots_adjust(left = 0.065, bottom=0.11, right=0.98, top=0.99, wspace= 0.25)
    
    
    for cls in range(len(class_names) - 1, -1, -1):
        for cluster in clusters:
            ax[1].add_patch(Rectangle((0, cluster[0] - 0.5), cluster[3][cls], cluster[1], facecolor = color_map[class_names[cls]], zorder = 10 + len(class_names) - cls))
    ax[1].set_xlim([0, max_bar_width * 1.1])
    ax[1].set_ylim([0, n])
    ax[1].spines['top'].set_visible(False)
    ax[1].spines['right'].set_visible(False)
    ax[1].yaxis.set_ticks_position('left')
    ax[1].xaxis.set_ticks_position('bottom')
    ax[1].get_yaxis().set_visible(False)
    ax[1].set_xlabel('Number of lipids per class')
    ax[1].legend(handles=[mpatches.Patch(color=color_map[key], label=key) for key in color_map], loc = cluster_components_legend_position, prop={'size': 8})
   
    

    matx = ax[0].matshow(sorted_correlations_np, cmap = colors_correlation)
    ax[0].set(xlim=[0, n], aspect = 1)
    for i, cluster in enumerate(clusters):
        ax[0].add_patch(Rectangle((cluster[0] - 0.5, cluster[0] - 0.5), cluster[1], cluster[1], facecolor = "none"))
        transFigure = fig.transFigure.inverted()
        y_pos = cluster[0] - 0.5 + cluster[1] / 2
        coord1 = transFigure.transform(ax[0].transData.transform([n, y_pos]))
        coord2 = transFigure.transform(ax[1].transData.transform([0, y_pos]))
        #coord3 = transFigure.transform(ax[2].transData.transform([fa_max, y_pos]))
        line = lines.Line2D((coord1[0],coord2[0]),(coord1[1],coord2[1]), transform = fig.transFigure, color = (0.4, 0.4, 0.4), linestyle = "dashed", label = "C%i" % (i + 1), zorder = 0)
        fig.lines += line,
        ax[0].text(coord1[0] + (1 if (i & 1) else 3) * (coord2[0] - coord1[0]) / 4, coord1[1], "C%i" % (i + 1), fontsize = 9, transform = fig.transFigure, horizontalalignment = "center", verticalalignment = "center", backgroundcolor = "white", zorder = 2)
        
    ax[0].set_xticklabels([])
    ax[0].get_yaxis().set_visible(False)
    ax[0].set_xlabel('Lipid-lipid Pearson correlations')
    transFigure = fig.transFigure.inverted()
    coords_origin = transFigure.transform(ax[0].transData.transform([0, 0]))
    cbar_axis = fig.add_axes([coords_origin[0] - 0.02, coords_origin[1], 0.01, 0.35])
    cbar = plt.colorbar(matx, ax=ax[0], ticks =[-1, 0, 1], cax = cbar_axis, label = "Correlation")
    cbar_axis.yaxis.set_ticks_position("left")
    cbar_axis.yaxis.set_label_position("left")
    
    
    plt.savefig(output_folder + folder_sep + "correlation_and_lipid_group.png", dpi = 300)
    plt.savefig(output_folder + folder_sep + "correlation_and_lipid_group.pdf", dpi = 300)
    fig.clf()


##################################################################################
# create lipid lipid fraction
##################################################################################



if create_lipid_lipid_fraction:
    time_stamp = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
    print("%s \t\t plotting lipid lipid fraction plot" % time_stamp)
    fig, ax = plt.subplots()
    fig.set_size_inches(8.6, 4)
    
    fractions = [[0, 0] for x in range(20)]
    for i in range(n - 1):
        for j in range(i + 1, n):
            c = correlations[i, j]
            index = floor((c + 1) * 10) if c < (1 - 1e-6) else 19
            fractions[index][0] += 1 if classes[i] == classes[j] else 0
            fractions[index][1] += 1
    
    bar_first = plt.bar(range(20), [1] * 20, 1, color=(0, 0.44, 0.67))
    bar_second = plt.bar(range(20), [fraction[0] / fraction[1] for fraction in fractions], 1, color=(0.6, 0, 0))
    x, w = np.arange(len(fractions)), 0.5
    
    
    plt.subplots_adjust(left=0.045, right=0.99, top=0.98, bottom=0.07)
    plt.xticks(x, ["-1", "-0.9", "-0.8", "-0.7", "-0.6", "-0.5", "-0.4", "-0.3", "-0.2", "-0.1", "0", "0.1", "0.2", "0.3", "0.4", "0.5", "0.6", "0.7", "0.8", "0.9", "1"])
    ax.legend((bar_first[0], bar_second[0]), ('between class', 'within class'))
    ax.set_xlabel('Lipid-lipid correlations')
    ax.set_ylabel('Fraction')
    
    plt.savefig(output_folder + folder_sep + "lipid_lipid_fraction.png", dpi = 300)
    plt.savefig(output_folder + folder_sep + "lipid_lipid_fraction.pdf", dpi = 300)


time_stamp = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
print("%s \t\t stored everything successfully in folder '%s' :-)" % (time_stamp, output_folder))