"""
MIT License

Copyright (c) 2017 Dominik Kopczynski   -   dominik.kopczynski {at} isas.de
                   Bing Peng   -   bing.peng {at} isas.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


####################################################################################
####### Please do NOT change anything, unless you know what you do!
####################################################################################

import random
import numpy
import matplotlib.pyplot as plt
import sys
from math import log
from lipid_parameters import *
import time
import datetime
import matplotlib.patches as mpatches
from itertools import repeat

if len(sys.argv) < 3:
    print("usage", sys.argv[0], "csv-file output-folder")
    exit()


input_csv = sys.argv[1]
output_folder = sys.argv[2]
folder_sep = "\\" if sys.platform[:3] == "win" else "/"

if create_violin_plot != True:
    exit()

time_stamp = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
print("%s \t\t reading '%s' as input file" % (time_stamp, input_csv))
row_data = {}
categories = {}






with open(input_csv, mode="rt") as infile:
    infile.readline() # first line only contains column names
    for line_i, line in enumerate(infile):
        tokens = line.strip().split(csv_delimiter)
        if len(tokens) < 2: continue
        key = tokens[1].split(" ")[0]
        if key not in row_data:
            row_data[key] = []
            if tokens[0] not in color_map:
                print("error in line %i, lipid category not registered in color map 'color_map'. Please consider the parameters section in the script" % (line_i + 1))
                exit()
            categories[key] = tokens[0]
        row_data[key].append(log(float(tokens[2]), 10))
        
time_stamp = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
print("%s \t\t input was successful, start creating the violin plot" % time_stamp)

del_key = []
for key in row_data:
    if len(row_data[key]) < 2: del_key.append(key)
    
for key in del_key:
    del row_data[key]
    del categories[key]
                 
pos = [i for i in range(len(row_data))]

data = []
data_keys = sorted(row_data.keys(), key = lambda x: categories[x] + str(numpy.mean(row_data[x])))
for key in data_keys:
    data.append(row_data[key])


x = numpy.arange(len(data))

ax = plt.gca()
ax.set_ylim([-1,len(data)])


violin_parts = plt.violinplot(data, pos, points=1000, vert=False, widths=0.7, showmeans=True, showextrema=True)
plt.yticks(x, data_keys)
plt.xticks([-4,-3,-2,-1, 0, 1, 2, 3, 4, 5, 6], [r'$10^{-4}$', r'$10^{-3}$', r'$10^{-2}$', r'$10^{-1}$', r'$10^0$', r'$10^1$', r'$10^2$', r'$10^3$', r'$10^4$', r'$10^5$', r'$10^6$'])

for pc, color in zip(violin_parts['bodies'], sorted(categories.keys(), key = lambda x: categories[x] + str(numpy.mean(row_data[x])))):
    
    pc.set_facecolor(color_map[categories[color]])
    pc.set_edgecolor('black')
    
fig = plt.gcf()
fig.set_size_inches(9, 12.7)
plt.subplots_adjust(left=0.14, right=0.98, top=0.99, bottom=0.03)
plt.legend(handles=[mpatches.Patch(color=color_map[key], label=key) for key in color_map], loc = violin_plot_legend_position, prop={'size': 10})

plt.savefig("%s%sviolinplot.png" % (output_folder, folder_sep), dpi=300)
plt.savefig("%s%sviolinplot.pdf" % (output_folder, folder_sep), dpi=300)

time_stamp = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
print("%s \t\t stored everything successfully in folder '%s' :-)" % (time_stamp, output_folder))
